package jobqueue

import (
	"fmt"
	"sync"
)

type Queue struct {
	jobMap              map[JobID]GenericJob
	pendingDependencies map[JobID][]JobID
	readyQueue          []GenericJob
	processedJobs       map[JobID]struct{}
	eventBus            *EventBus
	mu                  sync.Mutex
	manger              *Manager
}

func NewQueue(EventBus *EventBus) *Queue {
	return &Queue{
		jobMap:              make(map[JobID]GenericJob),
		pendingDependencies: make(map[JobID][]JobID),
		readyQueue:          []GenericJob{},
		processedJobs:       make(map[JobID]struct{}),
		eventBus:            EventBus,
	}
}

func (q *Queue) SetManager(m *Manager) {
	q.manger = m
}

func (q *Queue) Enqueue(job GenericJob) error {
	q.mu.Lock()
	defer q.mu.Unlock()

	if _, exists := q.jobMap[job.GetID()]; !exists {
		q.jobMap[job.GetID()] = job
	}

	for _, readyJob := range q.readyQueue {
		if readyJob.GetID() == job.GetID() {
			return ErrJobAlreadyExists
		}
	}

	// Check if this job is a dependency for any pending jobs
	for pendingJobID, pendingDeps := range q.pendingDependencies {
		q.pendingDependencies[pendingJobID] = removeJobID(pendingDeps, job.GetID())
		if len(q.pendingDependencies[pendingJobID]) == 0 {
			q.readyQueue = append(q.readyQueue, q.jobMap[pendingJobID])
			delete(q.pendingDependencies, pendingJobID)
		}
	}

	// Check this job's dependencies
	var unmetDependencies []JobID
	for _, depID := range job.GetDependencies() {
		if _, ok := q.jobMap[depID]; !ok {
			unmetDependencies = append(unmetDependencies, depID)
		}
	}

	if len(unmetDependencies) > 0 {
		q.pendingDependencies[job.GetID()] = unmetDependencies
	} else {
		q.readyQueue = append(q.readyQueue, job)

		// Run topological sort on jobs in the ready queue
		readyJobList := []GenericJob{}
		for _, readyJob := range q.readyQueue {
			readyJobList = append(readyJobList, readyJob)
		}

		for id := range q.processedJobs {
			readyJobList = append(readyJobList, q.jobMap[id])
		}

		sortedIDs, err := topologicalSortJobs(readyJobList)
		if err != nil {
			return err
		}

		// Reorder q.readyQueue based on sorted job IDs
		newReadyQueue := make([]GenericJob, len(sortedIDs))
		for i, id := range sortedIDs {
			newReadyQueue[i] = q.jobMap[id]
		}
		q.readyQueue = newReadyQueue

		if q.eventBus != nil && len(q.readyQueue) > 0 {
			if q.manger != nil && q.manger.logger != nil {
				q.manger.logger.Info("Job ready", "job_id", job.GetID())
			}

			q.eventBus.Publish(JobReady, job.GetID())
		}

	}

	return nil
}

func (q *Queue) Dequeue() (GenericJob, error) {
	q.mu.Lock()
	defer q.mu.Unlock()

	if len(q.readyQueue) == 0 {
		return nil, ErrQueueEmpty
	}

	job := q.readyQueue[0]
	q.readyQueue = q.readyQueue[1:]

	// Mark the job as processed but keep it in the jobMap for dependency resolution
	q.processedJobs[job.GetID()] = struct{}{}

	fmt.Printf("Dequeue: jobs in ready queue: %v, jobs in processed: %v, current job: %v\n", q.readyQueue, q.processedJobs, job.GetID())

	return job, nil
}

func removeJobID(deps []JobID, id JobID) []JobID {
	for i, dep := range deps {
		if dep == id {
			deps[i] = deps[len(deps)-1]
			return deps[:len(deps)-1]
		}
	}

	return deps
}
